# -*- coding: utf-8 -*-
from dateutil.parser import parse as parse_datetime


class Base(object):

    def prepare(self, value):
        return str(value) if value is not None else None

    def clean(self, text):
        if text is not None:
            return text


class String(Base):
    """String values [saml-core § 1.3.1].
    """


class Integer(Base):
    """Integral values.
    """


class Boolean(Base):
    """Boolean values.
    """

    def prepare(self, value):
        if value is None:
            return None

        return 'true' if value else 'false'

    def clean(self, text):
        if not text:
            return None

        return text == 'true'


class DateTime(Base):
    """
    An ISO 8601 formatted, UTC (eg. 2008-01-23T04:56:22Z)
    date-time [saml-core § 1.3.3].
    """

    @staticmethod
    def to_iso8601(when):
        text = when.strftime("%Y-%m-%dT%H:%M:%SZ")
        return text

    @staticmethod
    def from_iso8601(when):
        return parse_datetime(when)

    def prepare(self, value):
        return DateTime.to_iso8601(value) if value is not None else None

    def clean(self, text):
        if text is not None:
            return DateTime.from_iso8601(text)



# -*- coding: utf-8 -*-
import re


def _upcase_first_letter(s):
    return s[0].upper() + s[1:]


def pascalize(name):
    name = _upcase_first_letter(name.strip())
    pattern = r'[-_\s]+(.)?'
    name = re.sub(pattern, lambda m: m.groups()[0].upper() if m else '', name)
    return name


class classproperty(object):
    """Declares a read-only `property` that acts on the class object.
    """

    def __init__(self, getter):
        self.getter = getter

    def __get__(self, obj, cls):
        return self.getter(cls)



# -*- coding: utf-8 -*-
"""
A python interface to produce and consume Security Assertion Markup
Language (SAML) v2.0 messages.

See: https://www.oasis-open.org/standards#samlv2.0
"""
# Version of the library.
from ._version import __version__, __version_info__  # noqa

# Version of the SAML standard supported.
from .schema import VERSION as SAML_VERSION

from .signature import sign, verify
from . import client

VERSION = __version__

__all__ = [
    'VERSION',
    'SAML_VERSION',
    'sign',
    'verify',
    'client'
]



from saml.schema import utils


class TestPascalize:

    def test_normal(self):
        text = 'something'

        assert utils.pascalize(text) == 'Something'

    def test_underscore(self):
        text = 'some_thing'

        assert utils.pascalize(text) == 'SomeThing'

    def test_already(self):
        text = 'someThing'

        assert utils.pascalize(text) == 'SomeThing'

    def test_dash(self):
        text = 'some-thing'

        assert utils.pascalize(text) == 'SomeThing'

    def test_null(self):
        text = 'SomeThing'

        assert utils.pascalize(text) == 'SomeThing'

    def test_pascal(self):
        text = 'Some_thing'

        assert utils.pascalize(text) == 'SomeThing'



# -*- coding: utf-8 -*-

"""
Sign and verify signatures using the `python-xmlsec` library.

.. autofunction:: sign
.. autofunction:: verify
"""


def sign(xml, stream, password=None):
    """
    Sign an XML document with the given private key file. This will add a
    <Signature> element to the document.

    :param lxml.etree._Element xml: The document to sign
    :param file stream: The private key to sign the document with
    :param str password: The password used to access the private key

    :rtype: None

    Example usage:
    ::
        from saml import schema
        from lxml import etree

        document = schema.AuthenticationRequest()
        xml_document = document.serialize()
        with open('my_key_file.pem', 'r+') as stream:
            sign(xml_document, stream)

        print etree.tostring(xml_document)

    Produces the following XML document:

    .. code-block:: xml

        <samlp:AuthnRequest
            xmlns:samlp="urn:oasis:names:tc:SAML:2.0:protocol"
            xmlns:saml="urn:oasis:names:tc:SAML:2.0:assertion"
            Version="2.0" ID="_6087de0b111b44349a70ff40191a4c0c"
            IssueInstant="2015-03-16T21:06:39Z">
            <Signature xmlns="http://www.w3.org/2000/09/xmldsig#">
                <SignedInfo>
                    <CanonicalizationMethod
                        Algorithm="http://www.w3.org/2001/10/xml-exc-c14n#"/>
                        <SignatureMethod
                            Algorithm="http://www.w3.org/2000/
                            09/xmldsig#rsa-sha1"/>
                            <Reference>
                                <Transforms>
                                    <Transform
                                        Algorithm="http://www.w3.org/2000/
                                        09/xmldsig#enveloped-signature"/>
                                </Transforms>
                                <DigestMethod
                                    Algorithm="http://www.w3.org/2000/
                                    09/xmldsig#sha1"/>
                                    <DigestValue>
                                        94O1FOjRE4JQYVDqStkYzne9StQ=
                                    </DigestValue>
                            </Reference>
                </SignedInfo>
                <SignatureValue>
                    aFYRRjtB3bDyLLJzLZmsn0K4SXmOpFYJ+8R8D31VojgiF37FOElbE56UFbm8BAjn
                    l2AixrUGXP4djxoxxnfBD/reYw5yVuIVXlMxKec784nF2V4GyrfwJOKaNmlVPkq5
                    c8SI+EkKJ02mwiail0Zvjb9FzwvlYD+osMSXvJXVqnGHQDVFlhwbBRRVB6t44/M3
                    TzC4mLSVhuvcpsm4GTQSpGkHP7HvweKN/OTc0aTy8Kh/YUrImwnUCii+J0EW4nGg
                    71eZyq/IiSPnTD09WDHsWe3g29kpicZXqrQCWeLE2zfVKtyxxs7PyEmodH19jXyz
                    wh9hQ8t6PFO47Ros5aV0bw==
                </SignatureValue>
            </Signature>
        </samlp:AuthnRequest>
    """

    # Import xmlsec here to delay initializing the C library in
    # case we don't need it.
    import xmlsec

    # Resolve the SAML/2.0 element in question.
    from saml.schema.base import _element_registry
    element = _element_registry.get(xml.tag)

    # Create a signature template for RSA-SHA1 enveloped signature.
    signature_node = xmlsec.template.create(
        xml,
        xmlsec.Transform.EXCL_C14N,
        xmlsec.Transform.RSA_SHA1)

    # Add the <ds:Signature/> node to the document.
    xml.insert(element.meta.signature_index, signature_node)

    # Add the <ds:Reference/> node to the signature template.
    ref = xmlsec.template.add_reference(
        signature_node, xmlsec.Transform.SHA1)

    # Add the enveloped transform descriptor.
    xmlsec.template.add_transform(ref, xmlsec.Transform.ENVELOPED)

    # Create a digital signature context (no key manager is needed).
    ctx = xmlsec.SignatureContext()

    # Load private key.
    key = xmlsec.Key.from_memory(stream, xmlsec.KeyFormat.PEM, password)

    # Set the key on the context.
    ctx.key = key

    # Sign the template.
    ctx.sign(signature_node)


def verify(xml, stream):
    """
    Verify the signaure of an XML document with the given certificate.
    Returns `True` if the document is signed with a valid signature.
    Returns `False` if the document is not signed or if the signature is
    invalid.

    :param lxml.etree._Element xml: The document to sign
    :param file stream: The private key to sign the document with

    :rtype: Boolean
    """
    # Import xmlsec here to delay initializing the C library in
    # case we don't need it.
    import xmlsec

    # Find the <Signature/> node.
    signature_node = xmlsec.tree.find_node(xml, xmlsec.Node.SIGNATURE)
    if signature_node is None:
        # No `signature` node found; we cannot verify
        return False

    # Create a digital signature context (no key manager is needed).
    ctx = xmlsec.SignatureContext()

    # Register <Response/> and <Assertion/>
    ctx.register_id(xml)
    for assertion in xml.xpath("//*[local-name()='Assertion']"):
        ctx.register_id(assertion)

    # Load the public key.
    key = None
    for fmt in [
            xmlsec.KeyFormat.PEM,
            xmlsec.KeyFormat.CERT_PEM]:
        stream.seek(0)
        try:
            key = xmlsec.Key.from_memory(stream, fmt)
            break
        except ValueError:  
            # xmlsec now throws when it can't load the key
            pass

    # Set the key on the context.
    ctx.key = key

    # Verify the signature.
    try:
        ctx.verify(signature_node)

        return True

    except Exception:
        return False



# -*- coding: utf-8 -*-
from . import types, base, Element, saml


class Base(base.Base):

    class Meta:
        namespace = 'samlp', 'urn:oasis:names:tc:SAML:2.0:protocol'


class _Message(saml._Message, Base):
    """Contains common information found in most SAML/2.0 protocols.
    """

    # A URI reference indicating the address to which this request has
    # been sent.
    destination = base.Attribute(types.String)

    # Indicates whether or not (and under what conditions) consent has
    # been obtained from a principal in the sending of this request.
    consent = base.Attribute(types.String)


class NameIDPolicy(Base):
    """
    Tailors the name identifier in the subjects of assertions
    resulting from an <AuthnRequest>.
    """

    # Specifies the URI reference corresponding to a name identifier format
    # defined in this or another specification.
    format = base.Attribute(types.String)

    # Optionally specifies that the assertion subject's identifier be
    # returned (or created) in the namespace of a service provider
    # other than the requester, or in the namespace of an affiliation
    # group of service providers
    sp_name_qualifier = base.Attribute(types.String, name='SPNameQualifier')

    # A Boolean value used to indicate whether the identity provider is
    # allowed, in the course of fulfilling the request, to create a
    # new identifier to represent the principal.
    allow_create = base.Attribute(types.Boolean)


class RequestedAuthenticationContext(Base):

    class Meta:
        name = 'RequestedAuthnContext'

    # Specifies the comparison method used to evaluate the requested
    # context classes or statements, one of "exact", "minimum",
    # "maximum", or "better". The default is "exact".
    comparison = base.Attribute(types.String)

    # Specifies one or more URI references identifying authentication
    # context classes or declarations.
    reference = Element(saml.AuthenticationContextReference, collection=True)


class Protocol:
    """
    The available URIs that identifies a SAML protocol binding to be used when
    returning the <Response> message.
    """

    _PREFIX = 'urn:oasis:names:tc:SAML:2.0:bindings:'

    SOAP = '%sSOAP' % _PREFIX

    POAS = '%sPAOS' % _PREFIX

    REDIRECT = '%sHTTP-Redirect' % _PREFIX

    POST = '%sHTTP-POST' % _PREFIX

    ARTIFACT = '%sHTTP-Artifact' % _PREFIX

    URI = '%sURI' % _PREFIX


class AuthenticationRequest(_Message):
    """
    Create a SAML AuthnRequest
    ::
        from saml import schema
        from datetime import datetime

        document = schema.AuthenticationRequest()
        document.id = '11111111-2222-3333-4444-555555555555'
        document.issue_instant = datetime(2000, 1, 1)
        document.assertion_consumer_service_index = 0
        document.attribute_consuming_service_index = 0
        document.issuer = 'https://sp.example.com/SAML2'

        policy = schema.NameIDPolicy()
        policy.allow_create = True
        policy.format = schema.NameID.Format.TRANSIENT
        document.policy = policy

        print document.tostring()

    Produces the following XML document:

    .. code-block:: xml

        <samlp:AuthnRequest
            xmlns:samlp="urn:oasis:names:tc:SAML:2.0:protocol"
            xmlns:saml="urn:oasis:names:tc:SAML:2.0:assertion"
            Version="2.0"
            ID="11111111-2222-3333-4444-555555555555"
            IssueInstant="2000-01-01T00:00:00Z"
            AssertionConsumerServiceIndex="0"
            AttributeConsumingServiceIndex="0">
            <saml:Issuer>https://sp.example.com/SAML2</saml:Issuer>
            <samlp:NameIDPolicy
                Format="urn:oasis:names:tc:SAML:2.0:nameid-format:transient"
                AllowCreate="true"/>
        </samlp:AuthnRequest>
    """

    class Meta:
        name = 'AuthnRequest'

    # A Boolean value. If "true", the identity provider MUST authenticate
    # the presenter directly rather than rely on a previous security context.
    force_authn = base.Attribute(types.Boolean)

    # A Boolean value. If "true", the identity provider and the user agent
    # itself MUST NOT visibly take control of the user interface from the
    # requester and interact with the presenter in a noticeable fashion.
    is_passive = base.Attribute(types.Boolean)

    # A URI reference that identifies a SAML protocol binding to be used when
    # returning the <Response> message.
    protocol = base.Attribute(types.String, name='ProtocolBinding')

    # Indirectly identifies the location to which the <Response> message
    # should be returned to the requester.
    assertion_consumer_service_index = base.Attribute(types.Integer)

    # Specifies by value the location to which the <Response> message MUST
    # be returned to the requester.
    assertion_consumer_service_url = base.Attribute(
        types.String, name='AssertionConsumerServiceURL')

    # Indirectly identifies information associated with the requester
    # describing the SAML attributes the requester desires or requires to be
    # supplied by the identity provider in the <Response> message.
    attribute_consuming_service_index = base.Attribute(types.Integer)

    # Specifies the human-readable name of the requester for use by the
    # presenter's user agent or the identity provider.
    provider_name = base.Attribute(types.String)

    # Specifies the requested subject of the resulting assertion(s).
    subject = Element(saml.Subject)

    # Specifies constraints on the name identifier to be used to represent
    # the requested subject.
    policy = Element(NameIDPolicy)

    # Specifies the requirements, if any, that the requester places on
    # the authentication context that applies to the responding provider's
    # authentication of the presenter.
    requested_context = Element(RequestedAuthenticationContext)

    # Specifies a set of identity providers trusted by the requester to
    # authenticate the presenter, as well as limitations and context
    # related to proxying of the <AuthnRequest> message to subsequent identity
    # providers by the responder.
    # TODO: scoping = Element(Scoping)


class StatusCode(Base):

    # URI prefix for the values in this enumeration.
    _PREFIX = "urn:oasis:names:tc:SAML:2.0:status:"

    # The request succeeded.
    SUCCESS = "{}Success".format(_PREFIX)

    # Failure due to an error on the part of the requester.
    REQUESTER = "{}Requester".format(_PREFIX)

    # The version of the request message was incorrect.
    VERSION_MISMATCH = "{}VersionMismatch".format(_PREFIX)

    # The provider wasn't able to successfully authenticate the principal.
    AUTHENTICATION_FAILED = "{}AuthnFailed".format(_PREFIX)

    # Unexpected or invalid content was encountered.
    INVALID_ATTRIBUTE_NAME_OR_VALUE = "{}InvalidAttrNameOrValue".format(
        _PREFIX)

    # The responding provider cannot support the requested name ID policy.
    INVALID_NAME_ID_POLICY = "{}InvalidNameIDPolicy".format(_PREFIX)

    # The authentication context requirements cannot be met.
    NO_AUTHENTICATION_CONTEXT = "{}NoAuthnContext".format(_PREFIX)

    # None of the supported identity providers are available.
    NO_AVAILABLE_IDP = "{}NoAvailableIDP".format(_PREFIX)

    # The responding provider cannot authenticate the principal passively.
    NO_PASSIVE = "{}NoPassive".format(_PREFIX)

    # None of the identity providers are supported by the intermediary.
    NO_SUPPORTED_IDP = "{}NoSupportedIDP".format(_PREFIX)

    # Not able to propagate logout to all other session participants.
    PARTIAL_LOGOUT = "{}PartialLogout".format(_PREFIX)

    # Cannot authenticate directly and not permitted to proxy the request.
    PROXY_COUNT_EXCEEDED = "{}ProxyCountExceeded".format(_PREFIX)

    # Is able to process the request but has chosen not to respond.
    REQUEST_DENIED = "{}RequestDenied".format(_PREFIX)

    # The SAML responder or SAML authority does not support the request.
    REQUEST_UNSUPPORTED = "{}RequestUnsupported".format(_PREFIX)

    # Deprecated protocol version specified in the request.
    REQUEST_VERSION_DEPRECATED = "{}RequestVersionDeprecated".format(
        _PREFIX)

    # Protocol version specified in the request message is too low.
    REQUEST_VERSION_TOO_LOW = "{}RequestVersionTooHigh".format(_PREFIX)

    # Protocol version specified in the request message is too high.
    REQUEST_VERSION_TOO_HIGH = "{}RequestVersionTooLow".format(_PREFIX)

    # Resource value provided in the request message is invalid.
    RESOURCE_NOT_RECOGNIZED = "{}ResourceNotRecognized".format(_PREFIX)

    # The response message would contain more elements than able.
    TOO_MANY_RESPONSES = "{}TooManyResponses".format(_PREFIX)

    # base.Attribute from an unknown attribute profile.
    UNKNOWN_ATTR_PROFILE = "{}UnknownAttrProfile".format(_PREFIX)

    # The responder does not recognize the principal.
    UNKNOWN_PRINCIPAL = "{}UnknownPrincipal".format(_PREFIX)

    # The SAML responder cannot properly fulfill the request.
    UNSUPPORTED_BINDING = "{}UnsupportedBinding".format(_PREFIX)

    # The status code value. This attribute contains a URI reference.
    value = base.Attribute(types.String)

    # A subordinate status code that provides more specific information
    # on an error condition.
    # TODO: code = Element('self')


class Status(Base):

    # A code representing the status of the activity carried out in
    # response to the corresponding request.
    code = Element(StatusCode, required=True)


class StatusResponse(_Message):
    """Extends the common message type with a status element.
    """

    # A reference to the identifier of the request to which the
    # response corresponds, if any.
    in_response_to = base.Attribute(types.String)

    # A code representing the status of the corresponding request.
    status = Element(Status)


class Response(StatusResponse):
    """
    Create a SAML Response
    ::
        from saml import schema
        from datetime import datetime

        document = schema.Response()
        document.id = '11111111-1111-1111-1111-111111111111'
        document.in_response_to = '22222222-2222-2222-2222-222222222222'
        document.issue_instant = datetime(2000, 1, 1, 1)
        document.issuer = 'https://idp.example.org/SAML2'
        document.destination = 'https://sp.example.com/SAML2/SSO/POST'
        document.status.code.value = schema.StatusCode.SUCCESS

        # Create an assertion for the response.
        document.assertions = assertion = schema.Assertion()
        assertion.id = '33333333-3333-3333-3333-333333333333'
        assertion.issue_instant = datetime(2000, 1, 1, 2)
        assertion.issuer = 'https://idp.example.org/SAML2'

        # Create a subject.
        assertion.subject = schema.Subject()
        assertion.subject.principal = '44444444-4444-4444-4444-444444444444'
        assertion.subject.principal.format = schema.NameID.Format.TRANSIENT
        data = schema.SubjectConfirmationData()
        data.in_response_to = '22222222-2222-2222-2222-222222222222'
        data.not_on_or_after = datetime(2000, 1, 1, 1, 10)
        data.recipient = 'https://sp.example.com/SAML2/SSO/POST'
        confirmation = schema.SubjectConfirmation()
        confirmation.data = data
        assertion.subject.confirmation = confirmation

        # Create an authentication statement.
        statement = schema.AuthenticationStatement()
        assertion.statements.append(statement)
        statement.authn_instant = datetime(2000, 1, 1, 1, 3)
        statement.session_index = '33333333-3333-3333-3333-333333333333'
        reference = schema.AuthenticationContextReference
        statement.context.reference = reference.PASSWORD_PROTECTED_TRANSPORT

        # Create a authentication condition.
        assertion.conditions = conditions = schema.Conditions()
        conditions.not_before = datetime(2000, 1, 1, 1, 3)
        conditions.not_on_or_after = datetime(2000, 1, 1, 1, 9)
        condition = schema.AudienceRestriction()
        condition.audiences = 'https://sp.example.com/SAML2'
        conditions.condition = condition

        print document.tostring()

    Produces the following XML document:

    .. code-block:: xml

        <samlp:Response
            xmlns:samlp="urn:oasis:names:tc:SAML:2.0:protocol"
            xmlns:saml="urn:oasis:names:tc:SAML:2.0:assertion"
            Version="2.0"
            ID="11111111-1111-1111-1111-111111111111"
            IssueInstant="2000-01-01T01:00:00Z"
            Destination="https://sp.example.com/SAML2/SSO/POST"
            InResponseTo="22222222-2222-2222-2222-222222222222">
            <saml:Issuer>https://idp.example.org/SAML2</saml:Issuer>
            <samlp:Status>
                <samlp:StatusCode
                    Value="urn:oasis:names:tc:SAML:2.0:status:Success"/>
            </samlp:Status>
            <saml:Assertion
                Version="2.0"
                ID="33333333-3333-3333-3333-333333333333"
                IssueInstant="2000-01-01T02:00:00Z">
                <saml:Issuer>https://idp.example.org/SAML2</saml:Issuer>
                <saml:Subject>
                    <saml:NameID
                        Format=
                        "urn:oasis:names:tc:SAML:2.0:nameid-format:transient">
                        44444444-4444-4444-4444-444444444444
                    </saml:NameID>
                    <saml:SubjectConfirmation
                        Method="urn:oasis:names:tc:SAML:2.0:cm:bearer">
                        <saml:SubjectConfirmationData
                            NotOnOrAfter="2000-01-01T01:10:00Z"
                            Recipient="https://sp.example.com/SAML2/SSO/POST"
                            InResponseTo=
                            "22222222-2222-2222-2222-222222222222"/>
                    </saml:SubjectConfirmation>
                </saml:Subject>
                <saml:Conditions
                    NotBefore="2000-01-01T01:03:00Z"
                    NotOnOrAfter="2000-01-01T01:09:00Z">
                    <saml:AudienceRestriction>
                        <saml:Audience>
                            https://sp.example.com/SAML2
                        </saml:Audience>
                    </saml:AudienceRestriction>
                </saml:Conditions>
                <saml:AuthnStatement
                    AuthnInstant="2000-01-01T01:03:00Z"
                    SessionIndex="33333333-3333-3333-3333-333333333333">
                    <saml:AuthnContext>
                        <saml:AuthnContextClassRef>
                            urn:oasis:names:tc:SAML:2.0:ac:classes:
                            PasswordProtectedTransport
                        </saml:AuthnContextClassRef>
                    </saml:AuthnContext>
                </saml:AuthnStatement>
            </saml:Assertion>
        </samlp:Response>
    """

    # Specifies an assertion by value, or optionally an encrypted assertion
    # by value.
    assertions = Element(saml.Assertion, collection=True)


class Artifact(Base):
    """
    The artifact value that the requester received and now wishes to
    translate into the protocol message it represents.
    """


class ArtifactResolve(_Message):
    """
    Used to request that a SAML protocol message be returned in an
    <ArtifactResponse> message by specifying an artifact that represents
    the SAML protocol message.
    """

    # The artifact value that the requester received and now wishes to
    # translate into the protocol message it represents.
    artifact = Element(Artifact, required=True)


class ArtifactResponse(StatusResponse):
    """
    Responds to an <ArtifactResolve> request with an embedded message
    that was referenced by the given artifact.
    """

    # The embedded message.
    message = Element(_Message)


class SessionIndex(Base):
    """The identifier that indexes a session at the message recipient.
    """


class LogoutRequest(_Message):
    """
    Create a SAML LogoutRequest
    ::
        from saml import schema
        from datetime import datetime

        document = schema.LogoutRequest()
        document.id = '11111111-1111-1111-1111-111111111111'
        document.issue_instant = datetime(2000, 1, 1)
        document.issuer = 'https://idp.example.org/SAML2'
        document.destination = 'https://sp.example.org/SAML2/logout'
        document.principal = 'myemail@mydomain.com'
        document.principal.format = schema.NameID.Format.EMAIL
        document.principal.name_qualifier = 'https://idp.example.org/SAML2'
        document.session_index = 'SESSION-22222222-2222-2222-2222-222222222222'

        print document.tostring()

    Produces the following XML document:

    .. code-block:: xml

        <samlp:LogoutRequest
            xmlns:samlp="urn:oasis:names:tc:SAML:2.0:protocol"
            xmlns:saml="urn:oasis:names:tc:SAML:2.0:assertion"
            Version="2.0"
            ID="11111111-1111-1111-1111-111111111111"
            IssueInstant="2000-01-01T00:00:00Z"
            Destination="https://idphost/adfs/ls/">
            <saml:Issuer>https://idp.example.org/SAML2</saml:Issuer>
            <saml:NameID
                NameQualifier="https://idp.example.org/SAML2"
                Format="urn:oasis:names:tc:SAML:1.1:nameid-format:emailAddress">
                myemail@mydomain.com
            </saml:NameID>
            <samlp:SessionIndex>
                SESSION-22222222-2222-2222-2222-222222222222
            </samlp:SessionIndex>
        </samlp:LogoutRequest>
    """

    # The time at which the request expires, after which the recipient
    # may discard the message.
    not_on_or_after = base.Attribute(types.DateTime)

    # An indication of the reason for the logout, in the
    # form of a URI reference.
    reason = base.Attribute(types.String)

    # The identifier and associated attributes
    # (in plaintext or encrypted form) that specify the principal as
    # currently recognized by the identity and service providers prior
    # to this request.
    principal = Element(saml.NameID, required=True)

    # The identifier that indexes this session at the message recipient.
    session_index = Element(SessionIndex)


class LogoutResponse(StatusResponse):
    """
    Create a SAML LogoutResponse
    ::
        from saml import schema
        from datetime import datetime

        document = schema.LogoutResponse()
        document.id = '22222222-2222-2222-2222-222222222222'
        document.in_response_to = '11111111-1111-1111-1111-111111111111'
        document.issue_instant = datetime(2000, 1, 1)
        document.issuer = 'https://idp.example.org/SAML2'
        document.destination = 'https://sp.example.com/SAML2/SLO/POST'
        document.status.code.value = schema.StatusCode.SUCCESS

        print document.tostring()

    Produces the following XML document:

    .. code-block:: xml

        <samlp:LogoutResponse
            xmlns:samlp="urn:oasis:names:tc:SAML:2.0:protocol"
            xmlns:saml="urn:oasis:names:tc:SAML:2.0:assertion"
            Version="2.0"
            ID="22222222-2222-2222-2222-222222222222"
            IssueInstant="2000-01-01T00:00:00Z"
            Destination="https://sp.example.com/SAML2/SLO/POST"
            InResponseTo="11111111-1111-1111-1111-111111111111">
            <saml:Issuer>https://idp.example.org/SAML2</saml:Issuer>
            <samlp:Status>
                <samlp:StatusCode
                    Value="urn:oasis:names:tc:SAML:2.0:status:Success"/>
            </samlp:Status>
        </samlp:LogoutResponse>
    """



from saml import client
from test_schema import build_authentication_request_simple
from six.moves.urllib.parse import quote_plus


def test_relay_state():
    target = build_authentication_request_simple()
    state = 'http://localhost:8080/'
    uri, _ = client.send('http://localhost', target.serialize(), state)

    relay_state_part = 'RelayState=%s' % quote_plus(state)
    assert relay_state_part in uri



# -*- coding: utf-8 -*-
import base64
import zlib
import six
from six.moves.urllib.parse import urlencode, parse_qs, unquote_plus
from lxml import etree
from saml import schema
from saml.schema.base import _element_registry


def send(uri, message, relay_state=None, protocol='redirect'):

    # Determine the name of parameter.
    element = _element_registry.get(message.tag)
    if not element:
        raise ValueError('unknown element', message)

    if protocol == 'redirect':
        # For sending a message through redirection; we need
        # to encode the message (depending on what it is) in the URI
        # as SAMLRequest or SAMLResponse.
        name = 'SAMLRequest'
        if isinstance(element, schema.StatusResponse):
            name = 'SAMLResponse'

        # Serialize and encode the message.
        text = base64.b64encode(zlib.compress(etree.tostring(message))[2:-4])

        # Build the parameters.
        parameters = {name: text}
        if relay_state:
            parameters['RelayState'] = relay_state

        # Append the parameters on the uri and return.
        uri = '%s?%s' % (uri, urlencode(parameters))
        return uri, None

    raise ValueError('unknown protocol', protocol)


def _text(str_or_bytes):
    if isinstance(str_or_bytes, six.text_type):
        return str_or_bytes

    return str_or_bytes.decode()


def receive(method, query_string, body):
    # Determine the protocol used and pare the appropriate data.
    method = method.upper()
    if method == 'GET':
        data = parse_qs(_text(query_string))
        binding = 'artifact' if 'SAMLArtifact' in data else 'redirect'

    elif method == 'POST':
        data = parse_qs(_text(body))
        binding = 'post'

    else:
        # Unknown method used.
        return None

    if binding in ('redirect', 'post'):
        # Pull the text out of the query.
        encoded = data.get('SAMLResponse', data.get('SAMLRequest'))
        if not encoded:
            # No SAML message found.
            return None

        # Decode the text.
        text = base64.b64decode(encoded[0])
        if binding == "redirect":
            text = zlib.decompress(text, -15)

        # Parse the text into xml.
        message = etree.XML(text)

        # Get the relay state if present.
        relay_state = data.get('RelayState')
        if relay_state:
            relay_state = unquote_plus(relay_state[0])

        # Return the message and the relay state.
        return message, relay_state



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from setuptools import setup, find_packages

setup(
    name='saml_tra',
    version='0.9.1',
    description='A python interface to produce and consume Security '
                'Assertion Markup Language (SAML) v2.0 messages.',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.12'
    ],
    author='traneo',
    author_email='traneo@gmail.com',
    url='https://github.com/traneo/python-saml',
    packages=find_packages('.'),
    install_requires=(
        # Extensions to the standard Python datetime module.
        # Provides ability to easily parse ISO 8601 formatted dates.
        'python-dateutil',

        # lxml is the most feature-rich and easy-to-use library for
        # processing XML and HTML in the Python language.
        'lxml',

        # Python bindings for the XML Security Library.
        'xmlsec'
    ),
    extras_require={
        'test': (
            # Test runner.
            'pytest',

            # Ensure PEP8 conformance.
            'pytest-pep8',

            # Ensure PyFlakes conformance.
            'pytest-flakes',

            # Ensure test coverage.
            'pytest-cov',
        )
    }
)



# -*- coding: utf-8 -*-
from uuid import uuid4
from datetime import datetime
from . import types, base, VERSION
Element = base.Element


class Base(base.Base):

    class Meta:
        namespace = 'saml', 'urn:oasis:names:tc:SAML:2.0:assertion'


class BaseID(Base):
    """Provides an extension point for identifiers.
    """

    # The security or administrative domain that qualifies the name.
    name_qualifier = base.Attribute(types.String)

    # Further qualifies a name with the name of a service provider.
    sp_name_qualifier = base.Attribute(types.String, name="SPNameQualifier")


class NameID(BaseID):
    """
    Is the identifier used in various SAML assertion
    constructs [saml-core § 2.2.3].
    """

    class Format:
        # URI prefix for the values in this enumeration.
        _PREFIX = "urn:oasis:names:tc:SAML:"

        # The interpretation is left to individual implementations.
        UNSPECIFIED = '%s1.1:nameid-format:unspecified' % _PREFIX

        # Indicates a form of an email address.
        EMAIL = '%s1.1:nameid-format:emailAddress' % _PREFIX

        # Is in the form specified by the X.509 recommendation [XMLSig].
        X509 = '%s1.1:nameid-format:X509SubjectName' % _PREFIX

        # Is in the format of a Windows domain qualified name.
        WINDOWS = '%s1.1:nameid-format:WindowsDomainQualifiedName' % _PREFIX

        # Is in the form of a Kerberos principal name.
        KEREBOS = '%s2.0:nameid-format:kerberos' % _PREFIX

        # Is the identifier of an entity that provides SAML-based services.
        ENTITY = '%s2.0:nameid-format:entity' % _PREFIX

        # Is a persistent opaque identifier for a principal.
        PERSISTENT = '%s2.0:nameid-format:persistent' % _PREFIX

        # Is an identifier with transient semantics.
        TRANSIENT = '%s2.0:nameid-format:transient' % _PREFIX

    # A URI classification of string-based identifier information.
    format = base.Attribute(types.String)

    # A name identifier established by a service provider.
    sp_provided_id = base.Attribute(types.String, name="SPProvidedID")


class Issuer(NameID):
    """
    Represents the issuer of a SAML assertion or protocol
    message [saml-core § 2.2.5].
    """

    # If no Format value is provided, then the value ENTITY is in effect.
    format = base.Attribute(types.String)


class _Message(Base):
    """Contains common information found in most SAML/2.0 communications.
    """

    # The version of this message.
    version = base.Attribute(types.String, default=VERSION, required=True)

    # The identifier for this message.
    id = base.Attribute(types.String, name='ID', required=True,
                        default=lambda: '_%s' % uuid4().hex)

    # The time instant of issue in UTC.
    issue_instant = base.Attribute(types.DateTime, required=True,
                                   default=datetime.utcnow)

    # The SAML authority that is making the claim(s) in the message.
    issuer = Element(Issuer, required=True)


class Statement(Base):
    """An extension point for statements made about a subject.
    """


class AuthenticationContextReference(Statement):
    """
    A URI reference identifying an authentication context class that
    describes the authentication context declaration that follows.
    """

    class Meta:
        name = 'AuthnContextClassRef'

    # URI prefix for the values in this enumeration.
    _PREFIX = "urn:oasis:names:tc:SAML:2.0:ac:classes:"

    # Authenticated through the use of a provided IP address.
    INTERNET_PROTOCOL = "{}InternetProtocol".format(_PREFIX)

    # Authenticated through a provided IP address and username/password.
    INTERNET_PROTOCOL_PASSWORD = "{}InternetProtocolPassword".format(
        _PREFIX)

    # Authenticated using a Kerberos ticket.
    KERBEROS = "{}Kerberos".format(_PREFIX)

    # Reflects no mobile customer registration procedures and an
    # authentication of the mobile device without requiring explicit
    # end-user interaction.
    MOBILE_ONE_FACTOR_UNREGISTERED = (
        "{}MobileOneFactorUnregistered".format(_PREFIX))

    # Reflects no mobile customer registration procedures and a
    # two-factor based authentication, such as secure device and user PIN.
    MOBILE_TWO_FACTOR_UNREGISTERED = (
        "{}MobileTwoFactorUnregistered".format(_PREFIX))

    # Reflects mobile contract customer registration procedures and a
    # single factor authentication.
    MOBILE_ONE_FACTOR_CONTRACT = "{}MobileOneFactorContract".format(
        _PREFIX)

    # Reflects mobile contract customer registration procedures and a
    # two-factor based authentication.
    MOBILE_TWO_FACTOR_CONTRACT = "{}MobileTwoFactorContract".format(
        _PREFIX)

    # Authenticates through the presentation of a password over an
    # unprotected HTTP session.
    PASSWORD = "{}Password".format(_PREFIX)

    # Authentication through the presentation of a password over a
    # protected session.
    PASSWORD_PROTECTED_TRANSPORT = "{}PasswordProtectedTransport".format(
        _PREFIX)

    # A principal had authenticated to an authentication authority at
    # some point in the past using any authentication context supported
    # by that authentication authority.
    PREVIOUS_SESSION = "{}PreviousSession".format(_PREFIX)

    # The principal authenticated by means of a digital signature where
    # the key was validated as part of an X.509 Public Key Infrastructure.
    X509 = "{}X509".format(_PREFIX)

    # The principal authenticated by means of a digital signature where
    # key was validated as part of a PGP Public Key Infrastructure.
    PGP = "{}PGP".format(_PREFIX)

    # The principal authenticated by means of a digital signature where
    # the key was validated via an SPKI Infrastructure.
    SPKI = "{}SPKI".format(_PREFIX)

    # This context class indicates that the principal authenticated by
    # means of a digital signature according to the processing rules
    # specified in the XML Digital Signature specification [XMLSig].
    XMLDSIG = "{}XMLDSig".format(_PREFIX)

    # A principal authenticates to an authentication authority using a
    # smartcard.
    SMARTCARD = "{}Smartcard".format(_PREFIX)

    # A principal authenticates to an authentication authority through
    # a two-factor authentication mechanism using a smartcard with
    # enclosed private key and a PIN.
    SMARTCARD_PKI = "{}SmartcardPKI".format(_PREFIX)

    # A principal uses an X.509 certificate stored in software to
    # authenticate to the authentication authority.
    SOFTWARE_PKI = "{}SoftwarePKI".format(_PREFIX)

    # The principal authenticated via the provision of a fixed-line
    # telephone number, transported via a telephony protocol such as ADSL.
    TELEPHONY = "{}Telephony".format(_PREFIX)

    # The principal is "roaming" (perhaps using a phone card) and
    # authenticates via the means of the line number, a user suffix,
    # and a password element.
    NOMAD_TELEPHONY = "{}NomadTelephony".format(_PREFIX)

    # The principal authenticated via the provision of a fixed-line
    # telephone number and a user suffix, transported via a telephony
    # protocol such as ADSL.
    PERSONAL_TELEPHONY = "{}PersonalTelephony".format(_PREFIX)

    # Authenticated via the means of the line number,
    # a user suffix, and a password element.
    AUTHENTICATED_TELEPHONY = "{}AuthenticatedTelephony".format(_PREFIX)

    # Authenticated by means of Secure Remote Password s[RFC 2945].
    SECURE_REMOTE_PASSWORD = "{}SecureRemotePassword".format(_PREFIX)

    # The principal authenticated by means of a client certificate,
    # secured with the SSL/TLS transport.
    TLS_CLIENT = "{}TLSClient".format(_PREFIX)

    # A principal authenticates through a time synchronization token.
    TIME_SYNC_TOKEN = "{}TimeSyncToken".format(_PREFIX)

    # Indicates that the authentication means are unspecified.
    UNSPECIFIED = "{}Unspecified".format(_PREFIX)


class AuthenticationContext(Base):
    """Specifies the context of an authentication event.
    """

    class Meta:
        name = 'AuthnContext'

    # A URI reference identifying an authentication context class that
    # describes the authentication context declaration that follows.
    reference = Element(AuthenticationContextReference,
                        default=AuthenticationContextReference.UNSPECIFIED)

    # TODO: <AuthnContextDecl> or <AuthnContextDeclRef>
    # TODO: <AuthenticatingAuthority>


class AttributeValue(Base):
    pass


class Attribute(Base):

    name_ = base.Attribute(types.String, name="Name")

    name_format = base.Attribute(types.String)

    value = Element(AttributeValue)


class AttributeStatement(Statement):

    attributes = Element(Attribute, collection=True)


class SubjectLocality(Base):
    """
    Specifies the DNS domain name and IP address for the system from
    which the assertion subject was authenticated.
    """


class AuthenticationStatement(Statement):
    """
    Describes a statement by the SAML authority asserting that the assertion
    subject was authenticated by a particular means at a particular time.
    """

    class Meta:
        name = 'AuthnStatement'

    # Specifies the time at which the authentication took place.
    authn_instant = base.Attribute(types.DateTime, required=True,
                                   default=datetime.utcnow)

    # Specifies the index of a particular session between the principal
    # identified by the subject and the authenticating authority.
    session_index = base.Attribute(types.String)

    # Specifies a time instant at which the session between the principal
    # identified by the subject and the SAML authority issuing this
    # statement MUST be considered ended.
    session_not_on_or_after = base.Attribute(types.DateTime)

    # Specifies the DNS domain name and IP address for the system from which
    # the assertion subject was apparently authenticated.
    subject_locality = Element(SubjectLocality)

    # The context used by the authenticating authority up to and including
    # the authentication event that yielded this statement.
    context = Element(AuthenticationContext)


class SubjectConfirmationData(Base):
    """Specifies constraints on allowing a subject to be confirmed.
    """

    # TODO: <KeyInfoConfirmationDataType/>
    # TODO: Arbitrary attributes
    # TODO: Arbitrary elements

    # A time instant before which the subject cannot be confirmed.
    not_before = base.Attribute(types.DateTime)

    # A time instant at which the subject can no longer be confirmed.
    not_on_or_after = base.Attribute(types.DateTime)

    # URI specifying to which an attesting entity can present the assertion.
    recipient = base.Attribute(types.String)

    # ID of a SAML message to the entity can present the assertion to.
    in_response_to = base.Attribute(types.String)

    # The network address to which the saml entity can present the assertion.
    address = base.Attribute(types.String)


class SubjectConfirmation(Base):
    """
    Provides the means for a relying party to verify the
    correspondence of the subject of the assertion with the party with whom
    the relying party is communicating.
    """

    class Method:

        # URI namespace prefix of the values.
        _PREFIX = 'urn:oasis:names:tc:SAML:2.0:cm:'

        # The subject is confirmed by the indicated data.
        BEARER = '{}bearer'.format(_PREFIX)

        # The subject is confirmed by the holding of a key.
        HOLDER_OF_KEY = '{}holder-of-key'.format(_PREFIX)

    # URI reference that identifies a protocol to confirm the subject.
    method = base.Attribute(types.String, required=True, default=Method.BEARER)

    # Identifies the entity expected to satisfy the enclosed requirements.
    principal = Element(NameID)

    # Confirmation information and constraints.
    data = Element(SubjectConfirmationData)


class Subject(Base):
    """The principal that is the subject of all statements in an assertion.
    """

    # Identifies the subject.
    principal = Element(NameID)

    # Information that allows the subject to be confirmed. If more than one
    # subject confirmation is provided, then satisfying any one of them is
    # sufficient to confirm the subject for the purpose of applying
    # the assertion.
    confirmation = Element(SubjectConfirmation, collection=True)


class Condition(Base):
    """Serves as an extension point for new conditions.
    """


class Conditions(Base):
    """
    Defines the SAML constructs that place constraints on the
    acceptable use of SAML assertions.
    """

    # A time instant before which the subject cannot be confirmed.
    not_before = base.Attribute(types.DateTime)

    # A time instant at which the subject can no longer be confirmed.
    not_on_or_after = base.Attribute(types.DateTime)

    # Specifies that the assertion is addressed to a particular audience.
    condition = Element(Condition, collection=True)


class Audience(Base):
    """A URI reference that identifies an intended audience.
    """


class AudienceRestriction(Condition):
    """
    Specifies that the assertion is addressed to one or more
    specific audiences.
    """

    audiences = Element(Audience, collection=True)


class OneTimeUse(Condition):
    """
    Allows an authority to indicate that the information
    in the assertion is likely to change very soon and fresh information
    should be obtained for each use.
    """


class Assertion(_Message):
    """
    This type specifies the basic information that is common to
    all assertions [saml-core § 2.3.3].
    """

    # The subject of the statement(s) in the assertion.
    subject = Element(Subject)

    # Conditions that MUST be evaluated when assessing the validity
    # of and/or when using the assertion.
    conditions = Element(Conditions)

    # Additional information related to the assertion that assists
    # processing in certain situations but which MAY be ignored by
    # applications that do not understand the advice or do not wish to
    # make use of it.
    # TODO: <Advice/>

    # Statements that are being asserted about the included subject.
    statements = Element(Statement, collection=True)



# -*- coding: utf-8 -*-
from collections import OrderedDict
from lxml import etree
import six
from .utils import pascalize, classproperty


class Options(object):

    def __init__(self, meta, name, data, bases):
        """
        Initializes the options object and defaults configuration not
        specified.
        """

        # Name of the element in its serialized form.
        self.name = meta.get('name')
        if self.name is None:
            # Generate a name if none is provided.
            self.name = pascalize(name)

        # The namespace of the element.
        self.namespace = meta.get('namespace')

        # Index into the elements of where we stick the signature block.
        self.signature_index = meta.get('signature_index', 1)

# Element registry in order to lookup for deserialize.
_element_registry = {}


class Declarative(type):

    @classmethod
    def __prepare__(cls, name, bases):
        return OrderedDict()

    @classmethod
    def _gather_metadata(cls, metadata, bases):
        for base in bases:
            if isinstance(base, cls) and hasattr(base, 'Meta'):
                # Append metadata.
                metadata.append(getattr(base, 'Meta'))

                # Recurse.
                cls._gather_metadata(metadata, base.__bases__)

    @classmethod
    def _is_derived(cls, name, bases):
        for base in bases:
            if isinstance(base, cls):
                # This is some sort of derived resource; good.
                return True

        # This is not derived at all from Resource (eg. is base).
        return False

    @classmethod
    def _get_attributes_dict(cls, obj):
        return {n: getattr(obj, n) for n in dir(obj)}

    def __new__(cls, name, bases, attrs):
        # Only continue if we are dervied from declarative.
        if not cls._is_derived(name, bases):
            return super(Declarative, cls).__new__(
                cls, name, bases, attrs)

        # Gather the attributes of all options classes.
        # Start with the base configuration.
        metadata = {}

        # Expand the options class with the gathered metadata.
        base_meta = []
        cls._gather_metadata(base_meta, bases)

        # Apply the configuration from each class in the chain.
        for meta in base_meta:
            metadata.update(**cls._get_attributes_dict(meta))

        # Apply the configuration from the current class.
        cur_meta = {}
        if attrs.get('Meta'):
            cur_meta = cls._get_attributes_dict(attrs['Meta'])
            metadata.update(**cur_meta)

        # Gather and construct the options object.
        meta = attrs['meta'] = Options(metadata, name, cur_meta, base_meta)

        # Collect declared attributes.
        attrs['_items'] = OrderedDict()

        # Collect attributes from base classes.
        for base in bases:
            values = getattr(base, '_items', None)
            if values:
                attrs['_items'].update(values)

        # Collect attributes from current class.
        attrs_l = list(filter(lambda x: issubclass(type(x[1]), Component),
                              attrs.items()))
        attrs_l.sort(key=lambda x: x[1].creation_counter)
        for key, attr in attrs_l:
            # If name reference is null; default to camel-cased name.
            if attr._name is None:
                attr._name = pascalize(key)

            # Store attribute in dictionary.
            attrs['_items'][attr._name] = attr

        # Continue initialization.
        obj = super(Declarative, cls).__new__(cls, name, bases, attrs)

        # Add this element to the element registry.
        _element_registry[obj.name] = obj

        # Return the constructed element class.
        return obj


class Component(object):

    # Tracks each time this field is created; used to keep fields
    # in order
    creation_counter = 0

    def __init__(self, type_, name=None, required=False, default=None):
        # Name of the attribute in its serialized form.
        self._name = name

        # Underlying type of the attribute.
        self.type = type_

        # Whether the attribute is required or not.
        self.required = required

        # The default value for this attribute.
        self.default = default
        if not callable(default):
            # Normalize self.default to always be a callable.
            self.default = lambda: default

        # Adjust the creation counter, and save our local copy.
        self.creation_counter = Component.creation_counter
        Component.creation_counter += 1

    def __delete__(self, instance):
        if instance is not None:
            # Being accessed as an instance; use the instance state.
            if self._name in instance._state:
                del instance._state[self._name]
                return

        # Prevent deletion.
        raise TypeError("attribute can't be deleted")


class Element(Component):

    def __init__(self, type_, **kwargs):
        # If an element is a collection it is a list of elements.
        self.collection = kwargs.pop('collection', False)

        # Continue the initialization the base element.
        super(Element, self).__init__(type_, **kwargs)

    @property
    def name(self):
        # Return the namespaced name of the element.
        return '{%s}%s' % (self.type.meta.namespace[1], self._name)

    @property
    def namespace(self):
        # Return the namespace of the underlying type.
        return self.type.meta.namespace

    def prepare(self, instance):
        # Retrieve the value of this attribute from the instance.
        value = instance._state.get(self._name)
        if value is None and self.default:
            # No value; use the default callable.
            self.__set__(instance, self.default())
            value = instance._state.get(self._name)

        # Return the value.
        return value

    def deserialize(self, xml):
        return self.type.deserialize(xml)

    def __get__(self, instance, owner=None):
        if instance is not None:
            # Being accessed as an instance; use the instance state.
            value = instance._state.get(self._name)
            if value is None and self.collection:
                # No value and we need to be a collection of things.
                instance._state[self._name] = value = []

            if value is None:
                # Build a default one of ourself.
                instance._state[self._name] = value = self.type()

            # Return the value.
            return value

        # Return ourself.
        return self

    def __set__(self, instance, value):
        if instance is not None:
            if isinstance(value, str):
                # Value is just text; construct the type.
                value = self.type(value)

            # Being accessed as an instance; use the instance state.
            if self.collection:
                if self._name not in instance._state:
                    instance._state[self._name] = []

                instance._state[self._name].append(value)

            else:
                instance._state[self._name] = value

            return

        # Prevent assignment.
        raise TypeError("attribute can't be assigned")


class Attribute(Component):

    def __init__(self, type_, name=None, required=False, default=None):
        # Initialize the base element first.
        super(Attribute, self).__init__(
            type_, name=name, required=required, default=default)

        # Instantiate the type reference with no parameters.
        if isinstance(self.type, type):
            self.type = self.type()

    @property
    def name(self):
        return self._name

    def prepare(self, instance):
        # Retrieve the value of this attribute from the instance.
        value = instance._state.get(self.name)
        if value is None and self.default:
            # No value; use the default callable.
            value = self.default()

        # Run the value through the underyling type's preparation method.
        value = self.type.prepare(value)

        # Return the value.
        return value

    def clean(self, text):
        # Wipe off the passed text and squish it into a python object
        # if needed.
        return self.type.clean(text)

    def __get__(self, instance, owner=None):
        if instance is not None:
            # Being accessed as an instance; use the instance state.
            return instance._state.get(self._name)

        # Return ourself.
        return self

    def __set__(self, instance, value):
        if instance is not None:
            # Being accessed as an instance; use the instance state.
            instance._state[self._name] = value
            return

        # Prevent assignment.
        raise TypeError("attribute can't be assigned")


class Base(six.with_metaclass(Declarative)):

    def __init__(self, text=None, **kwargs):
        # Instance state of the attribute.
        self._state = {}

        # Text of the element.
        self.text = text

        # Update the instance state with kwargs.
        self._state.update(kwargs)

        # The signature function tuple.
        self._sign_args = None

    @classproperty
    def name(cls):
        # Return the namespaced name of the element.
        return '{%s}%s' % (cls.meta.namespace[1], cls.meta.name)

    def prepare(self):
        """Prepare the date in the instance state for serialization.
        """

        # Create a collection for the attributes and elements of
        # this instance.
        attributes, elements = OrderedDict(), []

        # Initialize the namespace map.
        nsmap = dict([self.meta.namespace])

        # Iterate through all declared items.
        for name, item in self._items.items():
            if isinstance(item, Attribute):
                # Prepare the item as an attribute.
                attributes[name] = item.prepare(self)

            elif isinstance(item, Element):
                # Update the nsmap.
                nsmap.update([item.namespace])

                # Prepare the item as an element.
                elements.append(item)

        # Return the collected attributes and elements
        return attributes, elements, nsmap

    def _serialize_item(self, item):
        # Destructure the data.
        attributes, elements, nsmap = item.prepare()

        # Create the XML node.
        node = etree.Element(item.name, nsmap=nsmap)

        # Add the attributes.
        for name, value in attributes.items():
            if value is not None:
                node.attrib[name] = value

        # Set its text.
        node.text = item.text

        # Iterate and serialize all elements.
        for element in elements:
            self._serialize_element(element, node)

        # Return the node.
        return node

    def _serialize_element(self, element, parent=None):
        # Prepare the instance state for serialization.
        items = element.prepare(self)
        if not items:
            # No data to serialize; move along.
            return

        try:
            # Serialize the item(s).
            for item in items:
                parent.append(item.serialize())

        except TypeError:
            # Serialize the single item.
            parent.append(items.serialize())

    def serialize(self):
        """
        Serializes the data in the instance state as an
        XML representation.
        """

        # Serialize the root and return the serialized element.
        return self._serialize_item(self)

    def tostring(self):
        return etree.tostring(self.serialize())

    @classmethod
    def deserialize(cls, xml):
        # Instantiate an instance of ourself.
        instance = cls()

        # Set the text element if present.
        if xml.text:
            instance.text = xml.text

        # Iterate through the items and deserialize them on the instance.
        elements = iter(xml.getchildren())
        element = None
        index = 0
        items = list(cls._items.values())
        # print(items)
        while index < len(items):
            # Fetch the next item.
            item = items[index]
            index += 1

            if isinstance(item, Attribute):
                # Attempt to get the attribute from the
                # xml element.
                value = xml.attrib.get(item.name)

                # Clean the value using the item clean.
                value = item.clean(value)

                # Set it on the instance.
                item.__set__(instance, value)

            elif isinstance(item, Element):
                if element is None:
                    try:
                        # Get the next element in the chain.
                        element = next(elements)

                    except StopIteration:
                        break

                # Resolve the element into a schema object.
                obj = _element_registry.get(element.tag)
                if obj is None:
                    # Element is unknown; bail.
                    element = None
                    index -= 1
                    continue

                # Is this element a subclass of the current item?
                if not issubclass(obj, item.type):
                    # Nope; skip to the next item.
                    continue

                # Deserialize the element.
                value = obj.deserialize(element)

                # Set it on the instance.
                item.__set__(instance, value)

                # Are we dealing with a "collection" ?
                if item.collection:
                    index -= 1

                # Unset the current element reference.
                element = None

        # Return the deserialized instance.
        return instance

    @classmethod
    def fromstring(cls, text):
        return cls.deserialize(etree.XML(text))



import saml
import xmlsec
from saml import schema
from saml.schema import utils
from datetime import datetime
from lxml import etree
from os import path
from pytest import mark

BASE_DIR = path.abspath(path.dirname(__file__))


def strip(text):
    if not text:
        return None

    text = text.replace('\n', '')
    text = text.strip()

    return text if text else None


def assert_node(expected, result):
    assert expected.tag == result.tag
    assert expected.attrib == result.attrib
    assert strip(expected.text) == strip(result.text)
    assert strip(expected.tail) == strip(result.tail)
    assert len(expected) == len(result)

    for expected, result in zip(expected, result):
        assert_node(expected, result)


def build_assertion_simple():
    # Create the assertion object.
    target = schema.Assertion()
    target.id = 'b07b804c-7c29-ea16-7300-4f3d6f7928ac'
    target.issue_instant = datetime(2004, 12, 5, 9, 22, 5)
    target.issuer = 'https://idp.example.org/SAML2'

    # Create a subject.
    target.subject = subject = schema.Subject()
    subject.principal = '3f7b3dcf-1674-4ecd-92c8-1544f346baf8'
    subject.principal.format = schema.NameID.Format.TRANSIENT
    subject.confirmation = confirmation = schema.SubjectConfirmation()
    confirmation.data = data = schema.SubjectConfirmationData()
    data.in_response_to = 'aaf23196-1773-2113-474a-fe114412ab72'
    data.not_on_or_after = datetime(2004, 12, 5, 9, 27, 5)
    data.recipient = 'https://sp.example.com/SAML2/SSO/POST'
    del data.recipient
    data.recipient = 'https://sp.example.com/SAML2/SSO/POST'

    assert data.recipient == 'https://sp.example.com/SAML2/SSO/POST'

    # Create an authentication statement.
    statement = schema.AuthenticationStatement()
    target.statements.append(statement)
    statement.authn_instant = datetime(2004, 12, 5, 9, 22, 0)
    statement.session_index = 'b07b804c-7c29-ea16-7300-4f3d6f7928ac'
    ref = schema.AuthenticationContextReference.PASSWORD_PROTECTED_TRANSPORT
    statement.context.reference = ref

    # Create a authentication condition.
    target.conditions = conditions = schema.Conditions()
    conditions.not_before = datetime(2004, 12, 5, 9, 17, 5)
    conditions.not_on_or_after = datetime(2004, 12, 5, 9, 27, 5)
    condition = schema.AudienceRestriction()
    condition.audiences = 'https://sp.example.com/SAML2'
    conditions.condition = condition

    # Return the built object.
    return target


def build_authentication_request_simple():
    # Create the authentication request.
    target = schema.AuthenticationRequest()
    target.id = 'aaf23196-1773-2113-474a-fe114412ab72'
    target.issue_instant = datetime(2004, 12, 5, 9, 21, 59)
    target.assertion_consumer_service_index = 0
    target.attribute_consuming_service_index = 0
    target.issuer = 'https://sp.example.com/SAML2'

    # Add a name id policy to the request.
    target.policy = policy = schema.NameIDPolicy()
    policy.allow_create = True
    policy.format = schema.NameID.Format.TRANSIENT

    # Return the built object.
    return target


def build_artifact_resolve_simple():
    # Create the artifact resolution request.
    target = schema.ArtifactResolve()
    target.id = '_cce4ee769ed970b501d680f697989d14'
    target.issue_instant = datetime(2004, 12, 5, 9, 21, 58)
    target.issuer = 'https://idp.example.org/SAML2'
    target.artifact = '''
        AAQAAMh48/1oXIM+sDo7Dh2qMp1HM4IF5DaRNmDj6RdUmllwn9jJHyEgIi8=
    '''.strip()

    # Return the built object.
    return target


def build_artifact_response_simple():
    # Create the artifact response.
    target = schema.ArtifactResponse()
    target.id = '_d84a49e5958803dedcff4c984c2b0d95'
    target.in_response_to = '_cce4ee769ed970b501d680f697989d14'
    target.issue_instant = datetime(2004, 12, 5, 9, 21, 59)
    target.status.code.value = schema.StatusCode.SUCCESS

    # Create an authentication request to stuff inside of the artifact
    # response.
    target.message = message = schema.AuthenticationRequest()
    message.id = '_306f8ec5b618f361c70b6ffb1480eade'
    message.issue_instant = datetime(2004, 12, 5, 9, 21, 59)
    message.destination = 'https://idp.example.org/SAML2/SSO/Artifact'
    message.protocol = schema.Protocol.ARTIFACT
    message.issuer = 'https://sp.example.com/SAML2'
    message.assertion_consumer_service_url = (
        'https://sp.example.com/SAML2/SSO/Artifact')

    # Add a name id policy to the authentication request.
    message.policy = policy = schema.NameIDPolicy()
    policy.allow_create = False
    policy.format = schema.NameID.Format.EMAIL

    # Return the built object.
    return target


def build_response_simple():
    # Create the response.
    target = schema.Response()
    target.id = 'identifier_2'
    target.in_response_to = 'identifier_1'
    target.issue_instant = datetime(2004, 12, 5, 9, 22, 5)
    target.issuer = 'https://idp.example.org/SAML2'
    target.destination = 'https://sp.example.com/SAML2/SSO/POST'
    target.status.code.value = schema.StatusCode.SUCCESS

    # Create an assertion for the response.
    target.assertions = assertion = schema.Assertion()
    assertion.id = 'identifier_3'
    assertion.issue_instant = datetime(2004, 12, 5, 9, 22, 5)
    assertion.issuer = 'https://idp.example.org/SAML2'

    # Create a subject.
    assertion.subject = subject = schema.Subject()
    subject.principal = '3f7b3dcf-1674-4ecd-92c8-1544f346baf8'
    subject.principal.format = schema.NameID.Format.TRANSIENT
    subject.confirmation = confirmation = schema.SubjectConfirmation()
    confirmation.data = data = schema.SubjectConfirmationData()
    data.in_response_to = 'identifier_1'
    data.not_on_or_after = datetime(2004, 12, 5, 9, 27, 5)
    data.recipient = 'https://sp.example.com/SAML2/SSO/POST'

    # Create an authentication statement.
    statement = schema.AuthenticationStatement()
    assertion.statements.append(statement)
    statement.authn_instant = datetime(2004, 12, 5, 9, 22, 0)
    statement.session_index = 'identifier_3'
    ref = schema.AuthenticationContextReference.PASSWORD_PROTECTED_TRANSPORT
    statement.context.reference = ref

    # Create a authentication condition.
    assertion.conditions = conditions = schema.Conditions()
    conditions.not_before = datetime(2004, 12, 5, 9, 17, 5)
    conditions.not_on_or_after = datetime(2004, 12, 5, 9, 27, 5)
    condition = schema.AudienceRestriction()
    condition.audiences = 'https://sp.example.com/SAML2'
    conditions.condition = condition

    # Return the built object.
    return target


def build_logout_request_simple():
    # Create the request.
    target = schema.LogoutRequest()
    target.id = 'identifier_1'
    target.issue_instant = datetime(2008, 6, 3, 12, 59, 57)
    target.issuer = 'myhost'
    target.destination = 'https://idphost/adfs/ls/'
    target.principal = 'myemail@mydomain.com'
    target.principal.format = schema.NameID.Format.EMAIL
    target.principal.name_qualifier = 'https://idphost/adfs/ls/'
    target.session_index = '_0628125f-7f95-42cc-ad8e-fde86ae90bbe'

    # Return the built object.
    return target


def build_logout_response_simple():
    # Create the response.
    target = schema.LogoutResponse()
    target.id = 'identifier_2'
    target.in_response_to = 'identifier_1'
    target.issue_instant = datetime(2004, 12, 5, 9, 22, 5)
    target.issuer = 'https://idp.example.org/SAML2'
    target.destination = 'https://sp.example.com/SAML2/SLO/POST'
    target.status.code.value = schema.StatusCode.SUCCESS

    # Return the built object.
    return target


NAMES = [
    'assertion',
    'authentication-request',
    'response',
    'logout-request',
    'logout-response',
    'artifact-resolve',
    'artifact-response'
]


@mark.parametrize('name', NAMES)
def test_simple_serialize(name):
    # Load the expected result.
    filename = path.join(BASE_DIR, '%s-simple.xml' % name)
    parser = etree.XMLParser(
        ns_clean=True, remove_blank_text=True, remove_comments=True)
    expected = etree.parse(filename, parser).getroot()

    # Build the result.
    build_fn_name = ('build-%s-simple' % name).replace('-', '_')
    target = globals()[build_fn_name]()

    # Serialize the result into an XML object.
    result = target.serialize()

    # Resolve and compare the result against the expected.
    assert_node(expected, result)


@mark.parametrize('name', NAMES)
def test_simple_deserialize(name):
    # Load the result.
    filename = path.join(BASE_DIR, '%s-simple.xml' % name)
    parser = etree.XMLParser(
        ns_clean=True, remove_blank_text=True, remove_comments=True)
    target = etree.parse(filename, parser).getroot()

    # Build the expected result.
    build_fn_name = ('build-%s-simple' % name).replace('-', '_')
    expected = globals()[build_fn_name]().serialize()

    # Deserialize and subsequently serialize the target.
    cls_name = utils.pascalize(name)
    result = getattr(schema, cls_name).deserialize(target).serialize()

    # Compare the nodes.
    assert_node(expected, result)


@mark.parametrize('name', NAMES)
def test_signed_deserialize(name):
    # Load the result.
    filename = path.join(BASE_DIR, '%s-signed.xml' % name)
    parser = etree.XMLParser(
        ns_clean=True, remove_blank_text=True, remove_comments=True)
    target = etree.parse(filename, parser).getroot()

    # Build the expected result.
    build_fn_name = ('build-%s-simple' % name).replace('-', '_')
    expected = globals()[build_fn_name]().serialize()

    # Deserialize and subsequently serialize the target.
    cls_name = utils.pascalize(name)
    result = getattr(schema, cls_name).deserialize(target).serialize()

    # TODO: Remove the signature node.

    # Compare the nodes.
    assert_node(expected, result)


@mark.parametrize('name', NAMES)
def test_generic_deserialize(name):
    filename = path.join(BASE_DIR, '%s-simple.xml' % name)
    parser = etree.XMLParser(
        ns_clean=True, remove_blank_text=True, remove_comments=True)
    target = etree.parse(filename, parser).getroot()

    build_fn_name = ('build-%s-simple' % name).replace('-', '_')
    expected = globals()[build_fn_name]().serialize()

    result = schema.deserialize(target).serialize()

    assert_node(expected, result)


def test_generic_deserialize_outside_registry():
    xml = build_authentication_request_simple().serialize()
    xml.tag = 'BadTagName'
    result = schema.deserialize(xml)

    assert result is None


@mark.parametrize('name', NAMES)
def test_sign(name):
    # Load the expected result.
    filename = path.join(BASE_DIR, '%s-signed.xml' % name)
    expected = etree.parse(filename).getroot()

    # Build the result.
    build_fn_name = ('build-%s-simple' % name).replace('-', '_')
    target = globals()[build_fn_name]()

    # Serialize the result into an XML object.
    # Round-trips the XML to remove weird spacing.
    result = target.serialize()

    with open(path.join(BASE_DIR, 'rsakey.pem'), 'r') as stream:
        # Sign the result.
        saml.sign(result, stream)

    # Compare the nodes.
    assert_node(expected, result)


@mark.parametrize('name', NAMES)
def test_verify(name):
    # Load the SAML XML document to verify.
    filename = path.join(BASE_DIR, '%s-signed.xml' % name)
    expected = etree.parse(filename).getroot()

    # Sign the result.
    with open(path.join(BASE_DIR, 'rsapub.pem'), 'r') as stream:
        assert saml.verify(expected, stream)


@mark.parametrize('name', NAMES)
def test_verify_with_bad_signature_returns_False(name):
    filename = path.join(BASE_DIR, '%s-signed.xml' % name)
    expected = etree.parse(filename).getroot()

    signature_node = xmlsec.tree.find_node(expected, xmlsec.Node.SIGNATURE)
    signature_node.clear()

    with open(path.join(BASE_DIR, 'rsapub.pem'), 'r') as stream:
        assert saml.verify(expected, stream) is False


@mark.parametrize('name', NAMES)
def test_verify_with_no_signature_returns_False(name):
    filename = path.join(BASE_DIR, '%s-simple.xml' % name)
    expected = etree.parse(filename).getroot()

    with open(path.join(BASE_DIR, 'rsapub.pem'), 'r') as stream:
        assert saml.verify(expected, stream) is False



# -*- coding: utf-8 -*-

"""
Create XML documents in accordance with the SAML 2.0 specification

AuthnRequest
------------
.. autoclass:: saml.schema.AuthenticationRequest

Response
--------
.. autoclass:: saml.schema.Response

LogoutRequest
-------------
.. autoclass:: saml.schema.LogoutRequest

LogoutResponse
--------------
.. autoclass:: saml.schema.LogoutResponse
"""

from .meta import version as VERSION  # noqa
from .saml import *  # noqa
from .samlp import *  # noqa


def deserialize(xml):
    from .base import _element_registry

    # Resolve the xml into an element.
    element = _element_registry.get(xml.tag)
    if not element:
        return None

    # Deserialize the xml and return.
    return element.deserialize(xml)
